/*
   Zoran.c

   ZORAN ZR36120 module driver.
   Copyright (C) 1997  Vincent Labie (try ivsparis@pratique.fr with
				      for Vincent Labie in the suject)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef		__MODULE__
#define		MODULE
#endif		//__MODULE__

#include	"Zoran.h"
#include	"ZoranProc.h"
#include	"ZoranIO.h"
#include	"SAA7110.h"

#define		PCI_VENDOR_ID_ZORAN		0x11de
#define		PCI_DEVICE_ID_ZORAN_ZR36120	0x6120
#define		PCI_ZORAN_LATENCY		32

int		zrDefault[]	={ZORAN_MAJOR,
				  ZORAN_ZRVERBOSE,ZORAN_SAAVERBOSE,ZORAN_I2CVERBOSE,
				  ZORAN_GENPURDIR,ZORAN_GENPURIO,ZORAN_GIRQ,
				  ZORAN_EXTFI,ZORAN_TOPFIELD,ZORAN_HSPOL,ZORAN_VSPOL};

/* module options zr=	<ZoranMajor>					*/
/* 			[<ZRVerbose>[<SAAVerbose>[<I2CVerbose>		*/
/*			[<GenPurDir>[<GenPurIO>[<GIRQ>			*/
/*			[<ExtFI>[<TopField>[<HSPol>[<VSPol>]]]]]]]]]]	*/
/* value>=256 means default						*/
/* ex insmod zr.o zr=256,1,1,0,256,256,256,0				*/
/*  => ZRVerbose=1,SAAVerbose=1,I2CVerbose=0,ExtFI=0,			*/
/*     all other=default						*/

int		zr[sizeof(zrDefault)/sizeof(int)]	={256,
							  256,256,256,
							  256,256,256,
							  256,256,256,256};

int		ZoranMajor;
extern	int	ZRVerbose;
extern	int	SAAVerbose;
extern	int	I2CVerbose;
extern	int	GenPurDir;
extern	int	GenPurIO;
extern	int	GIRQ;
extern	int	ExtFI;
extern	int	TopField;
extern	int	HSPol;
extern	int	VSPol;

int		NbZoran;
ZORAN *		TabZoran[NB_MAX_ZORAN_CARD];
u32		IrqMask=0;

	int	ZoranIOCTL2(ZORAN * pZoran,unsigned int Command,unsigned long arg);

/**********************************************************************************************************************/
	void	FreeFrameBuffer(ZORAN * pZoran)
{
	if (pZoran->pFrameBuffer!=NULL)
		{
		ZRDBG(("Zoran%d: freeing frame buffer address 0x%08X\n",
			pZoran->IndexPCI,(u32)pZoran->pFrameBuffer));
		bigphysarea_free(pZoran->pFrameBuffer,pZoran->FrameBufferSize);
		}
	pZoran->pFrameBuffer	=NULL;
	pZoran->FrameBufferSize	=0;
	pZoran->NbSlot		=0;
	pZoran->CapSlot		=-1;
	pZoran->UserSlot	=-1;
	memset(pZoran->TabSlotFilled,0,sizeof(pZoran->TabSlotFilled));
	memset(pZoran->TabSlotOffset,0,sizeof(pZoran->TabSlotOffset));
}

/**********************************************************************************************************************/
	BOOL	AllocFrameBuffer(ZORAN * pZoran,int NbSlot,u32 FrameSize)
{
	int			IndexSlot;
	int			FrameBufferSize=NbSlot*FrameSize;

	memset(pZoran->TabSlotFilled,0,sizeof(pZoran->TabSlotFilled));
	memset(pZoran->TabSlotOffset,0,sizeof(pZoran->TabSlotOffset));

	FrameBufferSize+=PAGE_SIZE;
	if (pZoran->FrameBufferSize!=FrameBufferSize)
		{
		if (pZoran->pFrameBuffer!=NULL)	FreeFrameBuffer(pZoran);
		pZoran->pFrameBuffer=(u8*)bigphysarea_alloc(FrameBufferSize);
		if (pZoran->pFrameBuffer==NULL)
			{ ERR(("Zoran%d: frame buffer allocation failed for %d bytes\n",pZoran->IndexPCI,FrameBufferSize));
			return(FALSE); }
		pZoran->FrameBufferSize=FrameBufferSize;
		}

	if (FrameSize==0)	pZoran->NbSlot=0;
	else			pZoran->NbSlot=NbSlot;
	ZRDBG(("Zoran%d: frame buffer allocated %i=FrameSize %i * NbSlot %i\n",
			pZoran->IndexPCI,FrameBufferSize,FrameSize,NbSlot));

	for(IndexSlot=0;IndexSlot<pZoran->NbSlot;IndexSlot++)
		{ pZoran->TabSlotOffset[IndexSlot]=FrameSize*IndexSlot;
		ZRDBG(("pZoran->TabSlotOffset[%i]=0x%08lX\n",
			IndexSlot,(long)pZoran->TabSlotOffset[IndexSlot])); }
	return(TRUE);
}

/**********************************************************************************************************************/
static	void ZoranSetLatency(const ZORAN * pZoran)
{ 
	u8			Latency;

	pcibios_read_config_byte(pZoran->Bus,pZoran->dfn,PCI_LATENCY_TIMER,&Latency);
	ZRDBG(("Zoran%d: (pZoran 0x%08X) read prev PCI_LATENCY_TIMER %i.\n",
		pZoran->IndexPCI,(u32)pZoran,(int)Latency));
	Latency=PCI_ZORAN_LATENCY;
	pcibios_write_config_byte(pZoran->Bus,pZoran->dfn,PCI_LATENCY_TIMER,Latency);
	ZRDBG(("Zoran%d: (pZoran 0x%08X) write PCI_LATENCY_TIMER %i.\n",
		pZoran->IndexPCI,(u32)pZoran,(int)Latency));
	pcibios_read_config_byte(pZoran->Bus,pZoran->dfn,PCI_LATENCY_TIMER,&Latency);
	ZRDBG(("Zoran%d: (pZoran 0x%08X) read back PCI_LATENCY_TIMER %i.\n",
		pZoran->IndexPCI,(u32)pZoran,(int)Latency));
}

/**********************************************************************************************************************/
static	int	ZoranInit(void)
{
	ZORAN *			pZoran;
	u16			IndexPCI;
	int			Result,FinalResult=0;
  
	if (pcibios_present()==FALSE)
		{ ERR(("ZR36120 driver: no PCI bios, aborting.\n")); return(-EUNATCH); }

	ERR(("ZR36120: driver version %i.%i\n",ZORAN_VERSION,ZORAN_SUBVERSION));
	for(IndexPCI=0;IndexPCI<NB_MAX_ZORAN_CARD;IndexPCI++)
		{
		u8		Bus,dfn;
		u8		Irq;
		void *		pZR36120;
		u32		ZR36120Page;
		u32		ZR36120Offset;
		u16		Command;

		Result=pcibios_find_device(PCI_VENDOR_ID_ZORAN,PCI_DEVICE_ID_ZORAN_ZR36120,
				IndexPCI,&Bus,&dfn);
		if (Result!=PCIBIOS_SUCCESSFUL) break;
		pZoran=(ZORAN*)kmalloc(sizeof(ZORAN),GFP_KERNEL);
		if (pZoran==NULL) { FinalResult=-ENOMEM; break; }

		memset(pZoran,0,sizeof(ZORAN)); TabZoran[IndexPCI]=pZoran;

		pZoran->Bus		=Bus;
		pZoran->dfn		=dfn;
		pZoran->IndexPCI	=IndexPCI;
		pZoran->bOpen		=FALSE;
		pZoran->FrameBufferSize	=0;
		pZoran->pFrameBuffer	=NULL;

		pcibios_read_config_byte(Bus,dfn,PCI_INTERRUPT_LINE,&Irq);
		pcibios_read_config_dword(Bus,dfn,PCI_BASE_ADDRESS_0,(void*)&pZR36120);

		pcibios_read_config_word(Bus,dfn,PCI_COMMAND,&Command);
		Command|=(PCI_COMMAND_MEMORY|PCI_COMMAND_MASTER);
		if (pcibios_write_config_word(Bus,dfn,PCI_COMMAND,Command)!=PCIBIOS_SUCCESSFUL)
			{ ERR(("Zoran%d: pcibios_write_config_word error\n",IndexPCI));
			FinalResult=-EFAULT; break; }

		ZoranSetLatency(pZoran);
		if ((IrqMask&(1<<Irq))==0)
			if (request_irq(Irq,ZoranIrqProc,(SA_INTERRUPT|SA_SHIRQ),"ZR36120",(void*)pZoran)==0)
				IrqMask|=(1<<Irq);
			else	{ ERR(("Zoran%d: can't allocate Irq %i\n",IndexPCI,Irq));
				FinalResult=-EBUSY; break; }
		pZoran->Irq=Irq;

		ZR36120Page	=((u32)pZR36120)&PAGE_MASK;
		ZR36120Offset	=((u32)pZR36120)&~PAGE_MASK;

		ZRDBG(("Prev ZR36120=0x%08X\n",(u32)pZR36120));
		pZR36120=(void*)(ZR36120Offset|
				(u32)vremap((u32)ZR36120Page,sizeof(ZR36120)));
		ZRDBG(("New ZR36120=0x%08X\n",(u32)pZR36120));
		if (pZR36120==NULL)	{ FinalResult=-EFAULT; break; }

		pZoran->pZR36120=pZR36120;
		ZoranReset(pZoran);
		ZRWAR((" Zoran%d: ",IndexPCI));	ZRDBG(("pZoran=0x%08X ",(u32)pZoran));
		ZRWAR(("Bus=%u ",(int)Bus));	ZRWAR(("dfn=%u ",(int)dfn));
		ZRWAR(("irq=%u ",(int)Irq));	ZRWAR(("IOMap=0x%08X ",(u32)pZR36120));
		ZRWAR(("\n"));
		}
	NbZoran=IndexPCI;
	if (FinalResult!=0)	ERR(("ZR36120: init failed.\n"));
	else			ERR(("ZR36120: init successfull.\n"));
	return(FinalResult);
}

/**********************************************************************************************************************/
static	int	ZoranRead(struct inode * inode,struct file * file,char * buffer,int count);
static	int	ZoranWrite(struct inode * inode,struct file * file,const char * buffer,int count);
static	int	ZoranIOCTL(struct inode *inode,struct file *file,unsigned int Command,unsigned long arg);
static	int	ZoranMMAP(struct inode * inode,struct file * file,struct vm_area_struct * vma);
static	int	ZoranOpen(struct inode * inode,struct file * file);
static	void	ZoranClose(struct inode * inode,struct file * file);

static	struct	file_operations		Zoran_fops=
	{
	NULL,		/* seek */
	ZoranRead,	/* read */
	ZoranWrite,	/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	ZoranIOCTL,	/* ioctl */
	ZoranMMAP,	/* mmap */
	ZoranOpen,	/* open */
	ZoranClose,	/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL		/* revalidate */
	};

/**********************************************************************************************************************/
#ifdef MODULE
	void	cleanup_module(void)
{  
	int		IndexPCI;

	for(IndexPCI=0;IndexPCI<NB_MAX_ZORAN_CARD;IndexPCI++ )
		{
		ZORAN *		pZoran=TabZoran[IndexPCI];
		if (pZoran==NULL) continue;
		if (pZoran->bOpen)
			{
			ERR(("ZR36120: GASP! cleanup module, Zoran%d in use. Crash in pending...\n",pZoran->IndexPCI));
			/*return;*/ // FIXME: don't know who to handle this!
			}
		}

	unregister_chrdev(ZoranMajor,"ZR36120");
	ZRWAR(("ZR36120 driver version %i.%i: ",ZORAN_VERSION,ZORAN_SUBVERSION));
	for(IndexPCI=0;IndexPCI<NB_MAX_ZORAN_CARD;IndexPCI++ )
		{
		ZORAN *		pZoran=TabZoran[IndexPCI];
		TabZoran[IndexPCI]=NULL;
		if (pZoran!=NULL)
			{
			if (pZoran->pZR36120!=NULL)
				{
				u16		Command;

				ZoranReset(pZoran);
				if (pcibios_read_config_word(pZoran->Bus,pZoran->dfn,
						PCI_COMMAND,&Command)==PCIBIOS_SUCCESSFUL)
					{
					Command&=~(PCI_COMMAND_MEMORY|PCI_COMMAND_MASTER);
					if (pcibios_write_config_word(pZoran->Bus,pZoran->dfn,
							PCI_COMMAND,Command)!=PCIBIOS_SUCCESSFUL)
						ERR(("Zoran%d: pcibios_write_config_word error\n",IndexPCI));
					}
//				vfree((void *)pZoran->pZR36120);	// FIXME hum,hum...?
				}
			if ((pZoran->Irq!=0)&&(IrqMask&(1<<(pZoran->Irq))))
				{ free_irq(pZoran->Irq,pZoran); IrqMask&=~(1<<pZoran->Irq);
				ZRDBG(("free irq %i, ",pZoran->Irq)); }
			FreeFrameBuffer(pZoran);
			kfree(pZoran);
			ZRWAR(("Zoran%d removed, ",IndexPCI));
			}
		}
	ZoranProcUnregister();
	ERR(("ZR36120: uninit successfull.\n"));
}

/**********************************************************************************************************************/
	int	init_module(void)
{
	int			Result		=0;
	int			IndexParam;
	int *			zrParams[sizeof(zrDefault)/sizeof(int)]
						={&ZoranMajor,
						  &ZRVerbose,&SAAVerbose,&I2CVerbose,
						  &GenPurDir,&GenPurIO,&GIRQ,
						  &ExtFI,&TopField,&HSPol,&VSPol};
	char *			zrParamsName[sizeof(zrDefault)/sizeof(int)]
						={"ZoranMajor",
						  "ZRVerbose","SAAVerbose","I2CVerbose",
						  "GenPurDir","GenPurIO","GIRQ",
						  "ExtFI","TopField","HSPol","VSPol"};
	int			zrParamsMaxVal[sizeof(zrDefault)/sizeof(int)]
						={254,
						  2,2,2,
						  255,255,1,
						  1,1,1,1};

	for(IndexParam=0;IndexParam<(sizeof(zrDefault)/sizeof(int));IndexParam++)
		if (zr[IndexParam]>=256) zr[IndexParam]=zrDefault[IndexParam];

	for(IndexParam=0;IndexParam<(sizeof(zrDefault)/sizeof(int));IndexParam++)
		if (zr[IndexParam]>zrParamsMaxVal[IndexParam])
			{ ERR(("ZR36120 driver: invalid value %i for %s, should not exceed %i\n",
				zr[IndexParam],zrParamsName[IndexParam],zrParamsMaxVal[IndexParam]));
			return(-EIO); }
		else	*zrParams[IndexParam]=zr[IndexParam];

	if (register_chrdev(ZoranMajor,"ZR36120",&Zoran_fops))
		{ ERR(("ZR36120 driver: can't get major %d for Zoran devices\n",ZoranMajor)); return(-EIO); }

	Result=ZoranInit();
	if (Result) { cleanup_module(); return(Result); }

	Result=ZoranProcRegister();
	if (Result) { cleanup_module(); return(Result); }

	return(0);
}

#endif /*MODULE*/

/**********************************************************************************************************************/
static ZORAN *	pZoranFromInode(struct inode *ip)
{
	int			minor=MINOR(ip->i_rdev);

	if (minor>=NbZoran) return(NULL);
	return(TabZoran[minor]);
}

/**********************************************************************************************************************/
static	int	ZoranOpen(struct inode * inode,struct file * file)
{
	ZORAN *			pZoran;

	pZoran=pZoranFromInode(inode); if (pZoran==NULL) return(-ENXIO);
	if (pZoran->bOpen) return(-EBUSY);

	pZoran->bOpen	=TRUE;
	pZoran->VideoStd=0;
	ZoranUnreset(pZoran);
	ZRDBG(("Zoran%d: (pZoran 0x%08X) open.\n",pZoran->IndexPCI,(u32)pZoran));
	return(0);
}

/**********************************************************************************************************************/
static	void	ZoranClose(struct inode * inode,struct file * file)
{
	ZORAN *			pZoran;

	pZoran=pZoranFromInode(inode); if (pZoran==NULL) return;
	if (pZoran->bOpen==FALSE)
		{ ERR(("Zoran%d: (pZoran 0x%08X) closed but not opened!.\n",
			pZoran->IndexPCI,(u32)pZoran)); return; }
	ZRDBG(("Zoran%d: (pZoran 0x%08X) close.\n",pZoran->IndexPCI,(u32)pZoran));
	pZoran->bOpen=FALSE;
	ZoranReset(pZoran);
	pZoran->bOpen=FALSE;
	udelay(10);
	FreeFrameBuffer(pZoran);
}

/**********************************************************************************************************************/
static	int	ZoranRead(struct inode * inode,struct file * file,char * buffer,int count)
{
	return(-EPERM);
}


/**********************************************************************************************************************/
static	int	ZoranWrite(struct inode * inode,struct file * file,const char * buffer,int count)
{
	return(-EPERM);
}

/**********************************************************************************************************************/
	int	ZoranIOCTL(struct inode *inode,struct file *file,unsigned int Command,unsigned long arg)
{
	ZORAN *			pZoran;
	pZoran=pZoranFromInode(inode); if (pZoran==NULL) return(-ENXIO);
	return(ZoranIOCTL2(pZoran,Command,arg));
}

/**********************************************************************************************************************/
static	int	ZoranMMAP(struct inode * inode,struct file * file,struct vm_area_struct * vma)
{
	ZORAN *			pZoran;
	int 			Error;
	u32			Size=vma->vm_end-vma->vm_start;
	u32			Offset;

	pZoran=pZoranFromInode(inode);	if (pZoran==NULL) return(-ENXIO);
	ZRDBG(("Zoran%d: (pZoran 0x%08X) mmap.\n",pZoran->IndexPCI,(u32)pZoran));
	Offset=(u32)pZoran->pFrameBuffer;
	if (pZoran->pFrameBuffer==NULL)
		{ ERR(("Zoran%d: (pZoran 0x%08X) no framebuffer to mmap.\n",
			pZoran->IndexPCI,(u32)pZoran));
		return(-ENOMEM); }
	if ((Offset|Size)&~PAGE_MASK)
		{ ERR(("Zoran%d: (pZoran 0x%08X) framebuffer to mmap is not page allined.\n",
			pZoran->IndexPCI,(u32)pZoran));
		return(-EFAULT); }
	if (Size>pZoran->FrameBufferSize)
		{ ERR(("Zoran%d: (pZoran 0x%08X) framebuffer to mmap is to big, ",
			pZoran->IndexPCI,(u32)pZoran));
		ERR(("Size=%i FrameBufferSize=%i.\n",Size,pZoran->FrameBufferSize));
		return(-EFBIG); }
	Error=remap_page_range(vma->vm_start,Offset,Size,vma->vm_page_prot);
	if (Error) { ERR(("Zoran%d: (pZoran 0x%08X) remap_page_range failed.\n",
				pZoran->IndexPCI,(u32)pZoran)); return(Error); }
	vma->vm_inode=inode;
	inode->i_count++;
	return(0);
}

